from functools import partial, wraps
from typing import Callable, Dict, Optional

from .constants import ENTRYPOINT_ATTR, PIPELINE_INPUT_ATTR, PIPELINE_NAME_ATTR
from .pfs import PFS
from .utils import in_container

__all__ = ("pipeline", "PFS")


def pipeline(
    pipeline_function: Optional[Callable] = None,
    *,
    name: Optional[str] = None,
    input: Optional[Dict] = None
):
    def set_attributes(func):
        setattr(func, ENTRYPOINT_ATTR, True)
        setattr(func, PIPELINE_NAME_ATTR, name)
        setattr(func, PIPELINE_INPUT_ATTR, input)

    if not in_container():
        # If we aren't running in the container, pass through
        if not pipeline_function:
            return partial(pipeline, name=name, input=input)
        set_attributes(pipeline_function)
        return pipeline_function

    def pipeline_decorator(pipeline_function_inner):
        @wraps(pipeline_function_inner)
        def wrapped_function(*args, **kwargs):
            return pipeline_function_inner(*args, **kwargs)
        set_attributes(wrapped_function)
        return wrapped_function

    if pipeline_function:
        return pipeline_decorator(pipeline_function)
    return pipeline_decorator



import os
import sys

from contextlib import redirect_stdout
from inspect import getmembers, isfunction
from importlib import import_module
from pathlib import Path
from subprocess import run
from tempfile import TemporaryDirectory
from types import ModuleType
from typing import Callable
from urllib.parse import urlparse

from .constants import ENTRYPOINT_ATTR, PIPELINE_ENV_VAR
from .models import PipelineSource


def in_container() -> bool:
    return bool(os.environ.get(PIPELINE_ENV_VAR))


def is_entrypoint(func: Callable) -> bool:
    return hasattr(func, ENTRYPOINT_ATTR)


def is_remote_location(location: str) -> bool:
    try:
        result = urlparse(location)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def nbconvert(notebook: bytes) -> bytes:
    command = [
        "jupyter-nbconvert",
        "--stdin",
        "--stdout",
        "--to=python",
        "--TagRemovePreprocessor.enabled=True",
        "--TagRemovePreprocessor.remove_cell_tags=pachyderm-ignore"
    ]
    result = run(command, capture_output=True, input=notebook)
    return result.stdout


def script_to_module(location: os.PathLike) -> PipelineSource:

    location = Path(location)
    original = source = location.read_bytes()
    if location.suffix == ".ipynb":
        source = nbconvert(original)

    module_name = "__pipeline_script"
    with TemporaryDirectory() as tempdir:
        test_script = Path(tempdir, f"{module_name}.py")
        test_script.write_bytes(source)

        sys.path.append(tempdir)
        try:
            with open(os.devnull, "w") as devnull:
                with redirect_stdout(devnull):
                    module = import_module(module_name)
            del sys.modules[module_name]
        except ImportError as err:
            raise RuntimeError(
                "Could not import the script. "
                "The script must be entirely self contained. "
                "Are all your dependencies installed? "
            ) from err
        assert sys.path.pop() == tempdir

    pipeline_source = PipelineSource(module, location, original, source)
    return pipeline_source


def get_entrypoint(module: ModuleType) -> Callable:
    """Returns the only function within the specified module marked with the
    @pipeline decorator.

    Args:
        module: The module to search.

    Raises:
        RuntimeError: If none or multiple entrypoints are found.

    Returns:
        Entrypoint function to the pipeline.
    """
    entrypoints = [
        func for _name, func in getmembers(module, isfunction)
        if is_entrypoint(func)
    ]
    if not entrypoints:
        raise RuntimeError(
            "No entrypoints found. "
            "Did you mark a function with the @pipeline decorator?"
        )
    if len(entrypoints) > 1:
        raise RuntimeError(
            "Multiple entrypoints found. "
            "Please only mark a single function with the @pipeline decorator."
        )
    return entrypoints[0]




ENTRYPOINT_ATTR = "__PACHYDERM_PIPELINE_ENTRYPOINT__"
PIPELINE_NAME_ATTR = "__PACHYDERM_PIPELINE_NAME__"
PIPELINE_INPUT_ATTR = "__PACHYDERM_PIPELINE_INPUTS__"

PIPELINE_ENV_VAR = "_PACHYDERM_PIPELINE_CONTAINER"
DEFAULT_TAG = "v0.1.0"
IMAGE = f"pachyderm/script_to_pipeline"



import sys

from base64 import b64decode
from importlib import import_module
from pathlib import Path
from tempfile import TemporaryDirectory

from script_to_pipeline.utils import get_entrypoint


if __name__ == "__main__":
    if len(sys.argv) > 1:

        source = b64decode(sys.argv[1])
        with TemporaryDirectory() as tempdir:
            Path(tempdir, "script.py").write_bytes(source)
            sys.path.append(tempdir)

            script = import_module("script")
            try:
                entrypoint = get_entrypoint(script)
            except RuntimeError:
                print("WARNING: No entrypoint found.")
            else:
                entrypoint()
    else:
        print("nothing to run")



from argparse import ArgumentParser
from pathlib import Path
from pprint import pprint
from subprocess import run

from ._internal import convert_to_specification, write_pipeline


def main() -> None:
    script, pipeline, update, image = parse_args()
    specification = convert_to_specification(script, image)
    if pipeline:
        write_pipeline(pipeline, specification)
        if update:
            run(["pachctl", "update", "pipeline", "-f", str(pipeline), "--reprocess"])
    else:
        pprint(specification, indent=2)


def parse_args():
    parser = ArgumentParser()
    parser.add_argument("script", type=Path, metavar="FILE")
    parser.add_argument("--pipeline", default=None, type=Path)
    parser.add_argument("--image", default=None, type=str)
    parser.add_argument("--update", action="store_true")
    args = parser.parse_args()
    if not args.script.exists():
        raise FileNotFoundError(args.script)
    return args.script, args.pipeline, args.update, args.image


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['script_to_pipeline', 'script_to_pipeline.entrypoint']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['script_to_pipeline = script_to_pipeline.__main__:main']}

setup_kwargs = {
    'name': 'script-to-pipeline',
    'version': '0.2.0a1',
    'description': 'Tool for converting python scripts into Pachyderm pipelines.',
    'long_description': None,
    'author': 'Ben Bonenfant',
    'author_email': 'bonenfan5ben@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path

from .utils import in_container


class PFS:

    @staticmethod
    def get(repo: str, out: bool = False) -> Path:
        if in_container():
            path = Path("/pfs", repo) if not out else Path("/pfs/out")
            if not path.exists():
                raise NotADirectoryError(path)
        else:
            raise NotImplementedError(
                "TODO: Add connection with mount extension for local development"
            )
        return path



import json
import os

from base64 import b64encode
from pathlib import Path
from typing import Dict

from .constants import PIPELINE_ENV_VAR, PIPELINE_INPUT_ATTR, PIPELINE_NAME_ATTR
from .utils import get_entrypoint, script_to_module


def convert_to_specification(script: os.PathLike, image: str) -> Dict:
    pipeline_source = script_to_module(script)
    entrypoint = get_entrypoint(pipeline_source.module)
    encoded_source = b64encode(pipeline_source.script)
    encoded_original = b64encode(pipeline_source.original)

    pipeline_name = getattr(entrypoint, PIPELINE_NAME_ATTR)
    name = dict(name=pipeline_name) if pipeline_name is not None else {}
    pipeline_input = getattr(entrypoint, PIPELINE_INPUT_ATTR) or dict()
    return dict(
        pipeline=name,
        input=pipeline_input,
        transform=dict(
            cmd=["python3", "-m", "script_to_pipeline.entrypoint", encoded_source.decode()],
            image=image,
            env={PIPELINE_ENV_VAR: "True"},
        ),
        metadata=dict(
            annotations=dict(
                encoded_source=encoded_original.decode(),
            ),
        )
    )


def write_pipeline(pipeline: os.PathLike, specification: Dict) -> None:
    pipeline = Path(pipeline)
    if not pipeline.exists():
        payload = dict(
            pipeline=dict(name=pipeline.stem),
            description="",
            input=dict(),
        )
    else:
        try:
            payload = json.loads(pipeline.read_bytes())
        except json.decoder.JSONDecodeError:
            raise RuntimeError(f"Could not read file: {pipeline}")
    if specification["pipeline"]:
        payload["pipeline"] = specification["pipeline"]
    if specification["input"]:
        payload["input"] = specification["input"]
    payload["transform"] = specification["transform"]
    payload["metadata"] = specification["metadata"]
    try:
        with pipeline.open("w") as file:
            json.dump(payload, file, indent=2)
    except IOError as err:
        raise RuntimeError(
            f"Could not create file: {pipeline}"
        ) from err
    return
